<?php

namespace Drupal\config_policy\Plugin\ConfigPolicyRule;

use Drupal\config_policy\Policy\ConfigPolicyRepositoryInterface;
use Drupal\config_policy\Result\FixResultInterface;
use Drupal\config_policy\Result\ResultItem\ErrorResultItem;
use Drupal\config_policy\Result\ResultItem\OkResultItem;
use Drupal\config_policy\Result\ResultItem\WarningResultItem;
use Drupal\config_policy\Result\ValidationResultInterface;
use Drupal\config_policy\Rule\ConfigRuleBase;
use Drupal\config_policy\Rule\FixableRuleInterface;
use Drupal\config_policy\Rule\PreventableRuleInterface;
use Drupal\config_policy\Rule\ValidatableRuleInterface;
use Drupal\Core\Config\Config;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Validates installed modules.
 *
 * @ConfigPolicyRule(
 *   id = "module_validation",
 *   label = @Translation("Module validation"),
 *   description = @Translation("Validates installed modules and can prevent them from being installed."),
 *   configPatterns = {
 *     "core.extension"
 *   },
 *   preventableForms = {
 *     "system_modules"
 *   }
 * )
 */
class ModuleValidation extends ConfigRuleBase implements ValidatableRuleInterface, FixableRuleInterface, PreventableRuleInterface {

  protected ModuleInstallerInterface $moduleInstaller;

  public function __construct(array $configuration, $plugin_id, $plugin_definition, LoggerInterface $logger, EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info, ConfigPolicyRepositoryInterface $config_policy_repository, ModuleInstallerInterface $module_installer) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $logger, $entity_type_manager, $entity_type_bundle_info, $config_policy_repository);
    $this->moduleInstaller = $module_installer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): ConfigRuleBase {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.factory')->get('config_policy'),
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('config_policy.repository'),
      $container->get('module_installer')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function validate(Config $config, ValidationResultInterface $result): ValidationResultInterface {
    $denied = $this->getInstalledDeniedModules($config);
    $advisory = $this->getInstalledAdvisoryModules($config);
    if (empty($denied)) {
      $result->add(new OkResultItem("No denied modules installed.", $config, $this));
    }

    foreach ($denied as $module) {
      $result->add(new ErrorResultItem("Disallowed module: $module", $config, $this));
    }

    if (empty($advisory)) {
      $result->add(new OkResultItem("No modules with an advisory installed.", $config, $this));
    }

    foreach ($advisory as $module) {
      $result->add(new WarningResultItem("Module with advisory: $module", $config, $this));
    }

    if (!empty($denied) || !empty($advisory)) {
      $result->addMessage($this->configuration['failure_message']);
    }

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function fix(Config $config, FixResultInterface $result): FixResultInterface {
    $denied = $this->getInstalledDeniedModules($config);
    foreach ($denied as $module) {
      $this->moduleInstaller->uninstall([$module]);
      $result->add(new OkResultItem("The module $module has been uninstalled.", $config, $this));
    }

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function prevent(array &$form, FormStateInterface $form_state, string $form_id): void {
    $denied = $this->getDeniedModules();

    foreach (Element::children($form['modules']) as $category) {
      foreach (Element::children($form['modules'][$category]) as $module) {
        if (in_array($module, $denied)) {
          $form['modules'][$category][$module]['enable']['#disabled'] = TRUE;
        }
      }
    }
  }

  protected function getDeniedModules(): array {
    return $this->configuration['denied'] ?? [];
  }

  protected function getAdvisoryModules(): array {
    return $this->configuration['advisory'] ?? [];
  }

  protected function getInstalledDeniedModules(Config $config): array {
    $installed = $config->get('module') ?? [];

    return array_intersect(
      array_keys($installed),
      $this->getDeniedModules()
    );
  }

  protected function getInstalledAdvisoryModules(Config $config): array {
    $installed = $config->get('module') ?? [];

    return array_intersect(
      array_keys($installed),
      $this->getAdvisoryModules()
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['denied'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Denied modules'),
      '#description' => $this->t('Enter the names of the modules to deny, separated by commas.'),
      '#default_value' => implode(', ', $this->configuration['denied'] ?? []),
      '#required' => TRUE,
    ];

    $form['advisory'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Modules with an advisory'),
      '#description' => $this->t('Enter the names of modules that have a negative advisory, these can still be installed.'),
      '#default_value' => implode(', ', $this->configuration['advisory'] ?? []),
      '#required' => TRUE,
    ];

    $form['failure_message'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Failure message'),
      '#description' => $this->t('An additional message to display upon failure of this rule.'),
      '#default_value' => $this->configuration['failure_message'] ?? '',
      '#required' => FALSE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::submitConfigurationForm($form, $form_state);

    $this->configuration['denied'] = array_map('trim', explode(',', $form_state->getValue('denied')));
    $this->configuration['advisory'] = array_map('trim', explode(',', $form_state->getValue('advisory')));
    $this->configuration['failure_message'] = $form_state->getValue('failure_message');
  }

}
